 /*************************************************************************
 ** This is the actual translation function. It does any necessary       **
 ** operations that are done after the translation, so all the main      **
 ** block needs to do is analyze args and determine the parameters to    **
 ** this method.                                                         **
 *************************************************************************/
func parse_file (program_file, debug_level, verbose_flag) {
  if (verbose_flag) putln ("& Entering parsing phase.");
  var token_vector,
  outside = except (). error (). invcall (). syserror(); // To save typing.
  /* The following function is called by the parser to
     do any error messaging. */
  func syntax_error (error_token_index,
                     error_token_value,
                     first_ignored_token_index,
                     first_ignored_token_value,
                     new_starting_token_index,
                     new_starting_token_value) {
    if (error_token_value == nil ||
        error_token_value.code != 31) {
      if (verbose_flag) putln ("! Syntax error.");
      if (error_token_value != nil)
        put (program_file, ":", error_token_value.line_num, ":0: ");
      error (sput ("Syntax error at token value ",
                        error_token_value
                        == nil ? "EOF" :
                        error_token_value.code,
                   ":", error_token_value
                        == nil ? "EOF" :
                        error_token_value.type0,
                   ":", error_token_value
                        == nil ? "EOF" :
                        sprint (error_token_value.value)
                  )
            );
    }
  }
  // The next few lines install the grammar.
  try { // In case we can't find the file.
    var tape = open (env {"GRAMMARPATH"},"read");
    grammar = fgetf (tape);
    close (tape);
  } catch (outside. eaccess,
           outside. eagain,
           outside. ebadf,
           outside. ebusy,
           outside. edom,
           outside. efault,
           outside. efbig,
           outside. eio,
           outside. eisdir,
           outside. emfile,
           outside. emlink,
           outside. enametoolong,
           outside. enfile,
           outside. enodev,
           outside. enoent,
           outside. enomem,
           outside. eperm,
           outside. epipe,
           outside. espipe,
           outside. exdev)
  {fatal_error ("Could not find grammar file 'grip.g'.");}
  var program, // Source code.
  abstract_tree; // Intermediate code.
  try { // To cause exception !pmemory to die with a graceful error message.
    var machine = parser();
    public machine;
    machine.set_grammar (grammar, 1);
    machine.set_debug (debug_level); // Turns on verbose mode if applicable.
    machine.set_one_parse (1); // In case the grammar is ambiguous.
    machine.set_recovery (1); // Turn off error recovery.
    machine.set_recovery_match (0); // Shift 5 tokens when error is found.
    try { // To cover 'no such file' errors.
      program = open (program_file, "read"); // Get the source code.
    } catch (outside. eaccess,
             outside. eagain,
             outside. ebadf,
             outside. ebusy,
             outside. edom,
             outside. efault,
             outside. efbig,
             outside. eio,
             outside. eisdir,
             outside. emfile,
             outside. emlink,
             outside. enametoolong,
             outside. enfile,
             outside. enodev,
             outside. enoent,
             outside. enomem,
             outside. eperm,
             outside. epipe,
             outside. espipe,
             outside. exdev
             ) { // No such file in reach.
      putln ("Error: No valid file named ",
             sprint (program_file), " found.");
      global_errfl++;
      abend (" I/O ");
    }
    source_code = fgetf (program) @ " ";
    // So we don't get out-of range array bounds during scanning.
    close(program);
    if (verbose_flag) putln ("@ Initiating scanner.");
    token_vector = scanner (source_code);
    if (verbose_flag && global_errfl)
      putln ("! Scanner failed, working in error recovery mode.");
    // Use scanner() to process source_code.
    if (verbose_flag) putln ("@ Initiating Earley's Parser.");
    abstract_tree =
    machine.parse (token_vector, syntax_error);
    // Parse token_vector into abstract_tree.
  } catch (except (). error (). invcall (). invparser (). pmemory) {
    // There wasn't enough memory to parse the program.
    if (verbose_flag)
      putln ("! Insufficient memory to set up parser, parsing failed.");
    error ("Insufficient memory for parser.");
  }
  if (global_errfl) { // If syntax errors were found, stop the program.
    if (verbose_flag) putln ("! Syntax error: Terminating now.");
    abend (" syntax ");
  } else if (verbose_flag) {
    putln ("& Parsing phase sucessful.");
  /* Here ends the grammar checking. abstract_tree contains the
     intermediate representation of the program. The task now is
     to translate it into an object of class prog, and check it
     for semantical correctness. */
  }
  include "anode_work"; // Used to derive an object of class prog().
  if (verbose_flag) print_anode (abstract_tree);
  return parse_program_anode (abstract_tree, verbose_flag);
}
